from abc import ABCMeta
from typing import Dict

from ortools.constraint_solver.pywrapcp import IntVar
from ortools.sat.python.cp_model import LinearExpr

from or_shifty.config import Config
from or_shifty.indexer import Idx
from or_shifty.shift import ShiftType


class Objective(metaclass=ABCMeta):
    def objective(self, assignments: Dict[Idx, IntVar], data: Config) -> LinearExpr:
        pass


class RankingWeight(Objective):
    ADDITIONAL_SHIFTS_WEIGHT = 100
    RANKING_WEIGHT = 1

    def objective(self, assignments: Dict[Idx, IntVar], data: Config) -> LinearExpr:
        # Compute the ranking weight for each shift_type. The coefficient is to discourage optimising one
        # shift type at the expense of another
        return LinearExpr.Sum(
            self._ranking_weight_for_shift_type(assignments, data, shift_type)
            for shift_type in ShiftType
        )

    def _ranking_weight_for_shift_type(
        self, assignments: Dict[Idx, IntVar], data: Config, shift_type: ShiftType
    ) -> LinearExpr:
        people_ranking = self._rank_people(data, shift_type)

        weights = self._assign_weights(data, people_ranking)

        # Multiply each person/person_shift assignment to a day_shift of the selected shift_type
        # (which should be either 0 or 1) with the corresponding weight and return the sum
        expressions = []
        coefficients = []
        for (person, person_shift), weight in weights.items():
            indices_to_include = [
                index
                for index in data.indexer.iter(
                    person_filter=person, person_shift_filter=person_shift
                )
                if index.day_shift.shift_type is shift_type
            ]
            expressions.append(
                sum(assignments[index.idx] for index in indices_to_include)
            )
            coefficients.append(weight)

        return LinearExpr.ScalProd(expressions, coefficients)

    def _rank_people(self, data, shift_type):
        # Rank people based on
        # 1. how many shifts they have done (DESC)
        # 2. how recently they have been on shift (ASC)
        # 3. the person's name to ensure there is a total ordering
        people = list(data.shifts_by_person.keys())
        historical_shifts = {
            person: data.history_metrics.num_of_shifts.get(shift_type, {}).get(
                person, 0
            )
            for person in people
        }
        free_days_since_last = {
            person: data.now - data.history_metrics.date_last_on_shift[person]
            for person in people
        }
        return sorted(
            people,
            key=lambda p: (historical_shifts[p], -free_days_since_last[p], p.name,),
            reverse=True,
        )

    def _assign_weights(self, data, people_ranking):
        shift_change_marker = object()

        # Rank people shifts s.t. everyone's first shift is ranked higher than anyone's second shift and so on
        people_shifts_ranking = []
        for person_shift_idx in reversed(range(data.max_shifts_per_person)):
            for person in people_ranking:
                people_shifts_ranking.append(
                    (person, data.shifts_by_person[person][person_shift_idx])
                )
            people_shifts_ranking.append(shift_change_marker)

        # Assign a weight to every person/shift to encourage assigning to pairs higher up in the ranking
        # first. The additional shifts weight when changing shifts ensures that we penalise more heavily
        # assigning someone to a second or third shift
        next_weight = 0
        weights = {}
        for entry in people_shifts_ranking:
            if entry is shift_change_marker:
                next_weight += self.ADDITIONAL_SHIFTS_WEIGHT
                continue

            weights[entry] = next_weight
            next_weight += self.RANKING_WEIGHT

        return weights


OBJECTIVE_FUNCTIONS = {objective.__name__: objective for objective in (RankingWeight,)}



from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum, auto
from typing import Any, Dict

from or_shifty.person import Person


@dataclass(frozen=True)
class Shift:
    name: str
    shift_type: "ShiftType"
    day: date

    @classmethod
    def from_json(cls, serialised: Dict[str, Any]) -> "Shift":
        return Shift(
            name=serialised["name"],
            shift_type=ShiftType.from_json(serialised["type"]),
            day=datetime.fromisoformat(serialised["day"]).date(),
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "type": self.shift_type.to_json(),
            "day": self.day.isoformat(),
        }

    def assign(self, person: Person) -> "AssignedShift":
        return AssignedShift(
            name=self.name, shift_type=self.shift_type, day=self.day, person=person,
        )


@dataclass(frozen=True)
class AssignedShift(Shift):
    person: Person

    def __str__(self):
        st_len = max(len(st.name) for st in ShiftType)
        shift_type = self.shift_type.name.lower()
        return f"{str(self.day)} ({self.name} - {shift_type:<{st_len}}) - {self.person.name}"

    @classmethod
    def from_json(cls, serialised: Dict[str, Any]) -> "AssignedShift":
        return AssignedShift(
            name=serialised["name"],
            shift_type=ShiftType.from_json(serialised["type"]),
            day=datetime.fromisoformat(serialised["day"]).date(),
            person=Person(name=serialised["person"]),
        )

    def to_json(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "type": self.shift_type.to_json(),
            "day": self.day.isoformat(),
            "person": self.person.name,
        }

    def unassigned(self) -> Shift:
        return Shift(name=self.name, shift_type=self.shift_type, day=self.day)


class ShiftType(Enum):
    STANDARD = auto()
    SPECIAL_A = auto()
    SPECIAL_B = auto()

    @classmethod
    def from_day(cls, day: date):
        if day.weekday() == 5:
            return ShiftType.SATURDAY
        if day.weekday() == 6:
            return ShiftType.SUNDAY
        return ShiftType.WEEKDAY

    def is_of_type(self, day: date, bank_holidays: Dict[date, "ShiftType"]):
        if day in bank_holidays:
            return self is bank_holidays[day]
        if day.weekday() == 5:
            return self is ShiftType.SATURDAY
        if day.weekday() == 6:
            return self is ShiftType.SUNDAY
        return self is ShiftType.WEEKDAY

    @classmethod
    def from_json(cls, serialised) -> "ShiftType":
        serialised = serialised.lower()
        if serialised == "standard":
            return ShiftType.STANDARD
        elif serialised == "special_a":
            return ShiftType.SPECIAL_A
        elif serialised == "special_b":
            return ShiftType.SPECIAL_B
        else:
            assert False, f"Unknown shift type {serialised}"

    def to_json(self) -> str:
        return self.name.lower()



from dataclasses import dataclass
from datetime import date
from typing import Dict, List

from or_shifty.history import History
from or_shifty.person import Person
from or_shifty.shift import ShiftType

NEVER = date(1970, 1, 1)


@dataclass(frozen=True)
class HistoryMetrics:
    num_of_shifts: Dict[ShiftType, Dict[Person, int]]
    date_last_on_shift: Dict[Person, date]
    date_last_on_shift_of_type: Dict[Person, Dict[ShiftType, date]]
    now: date

    @classmethod
    def build(cls, history: History, people: List[Person], now: date):
        return cls(
            num_of_shifts=cls._num_of_shifts(history, people),
            date_last_on_shift=cls._date_last_on_shift(history, people),
            date_last_on_shift_of_type=cls._date_last_on_shift_of_type(history, people),
            now=now,
        )

    @classmethod
    def _num_of_shifts(cls, history: History, people: List[Person]):
        return {
            shift_type: cls._num_of_shifts_for_type(history, people, shift_type)
            for shift_type in ShiftType
        }

    @classmethod
    def _num_of_shifts_for_type(
        cls, history: History, people: List[Person], shift_type: ShiftType
    ):
        shifts = {person: 0 for person in people}

        for offset in history.offsets:
            if offset.shift_type is shift_type:
                shifts[offset.person] = offset.offset

        for past_shift in history.past_shifts:
            if past_shift.shift_type is shift_type and past_shift.person in shifts:
                shifts.setdefault(past_shift.person, 0)
                shifts[past_shift.person] += 1

        return shifts

    @classmethod
    def _date_last_on_shift(cls, history: History, people: List[Person]):
        people_seen = set()
        date_last = {person: NEVER for person in people}
        for past_shift in history.past_shifts:
            if past_shift.person in date_last and past_shift.person not in people_seen:
                people_seen.add(past_shift.person)
                date_last[past_shift.person] = past_shift.day
        return date_last

    @classmethod
    def _date_last_on_shift_of_type(cls, history: History, people: List[Person]):
        people_seen = {person: set() for person in people}
        date_last = {
            person: {shift_type: NEVER for shift_type in ShiftType} for person in people
        }
        for past_shift in history.past_shifts:
            if (
                past_shift.person in date_last
                and past_shift.shift_type not in people_seen[past_shift.person]
            ):
                people_seen[past_shift.person].add(past_shift.shift_type)
                date_last[past_shift.person][past_shift.shift_type] = past_shift.day
        return date_last

    def __str__(self):
        formatted = "Pre-allocation history metrics:\n"
        formatted += "{: <20}{: <15}{: <15}{: <15}{: <15}\n".format(
            "Name", "Standard", "Special A", "Special B", "Last on"
        )

        for person in self.date_last_on_shift.keys():
            if len(person.name) > 16:
                formatted += f"{person.name[:16] + '...': <20}"
            else:
                formatted += f"{person.name: <20}"

            formatted += f"{self.num_of_shifts[ShiftType.STANDARD][person]: <15}"
            formatted += f"{self.num_of_shifts[ShiftType.SPECIAL_A][person]: <15}"
            formatted += f"{self.num_of_shifts[ShiftType.SPECIAL_B][person]: <15}"
            formatted += f"{(self.date_last_on_shift[person] - self.now).days: <15}"

            formatted += "\n"
        return formatted



from dataclasses import dataclass
from datetime import date
from typing import Dict, Generator, List, Tuple

from or_shifty.person import Person
from or_shifty.shift import Shift

Idx = Tuple[int, int, int, int]
PersonShift = int


@dataclass(frozen=True)
class IndexEntry:
    idx: Idx
    person: Person
    person_shift: PersonShift
    day: date
    day_shift: Shift


@dataclass(frozen=True)
class Indexer:
    _person_indices: Dict[Person, int]
    _day_indices: Dict[date, int]
    _day_shift_indices: Dict[Tuple[date, Shift], int]

    _index_entries: Dict[Idx, IndexEntry]

    @classmethod
    def build(
        cls,
        people: List[Person],
        max_shifts_per_person: int,
        shifts_per_day: Dict[date, List[Shift]],
    ):
        _person_indices = {}
        _day_indices = {}
        _day_shift_indices = {}
        _index_entries = {}

        for person_idx, person in enumerate(people):
            _person_indices[person] = person_idx

            for person_shift_idx in range(max_shifts_per_person):
                for day_idx, day in enumerate(sorted(shifts_per_day.keys())):
                    _day_indices[day] = day_idx

                    for shift_idx, shift in enumerate(
                        sorted(shifts_per_day[day], key=lambda s: s.name)
                    ):
                        _day_shift_indices[(day, shift)] = shift_idx

                        idx = (
                            person_idx,
                            person_shift_idx,
                            day_idx,
                            shift_idx,
                        )
                        _index_entries[idx] = IndexEntry(
                            idx, person, person_shift_idx, day, shift,
                        )

        return cls(
            _person_indices=_person_indices,
            _day_indices=_day_indices,
            _day_shift_indices=_day_shift_indices,
            _index_entries=_index_entries,
        )

    def get(self, index: Idx) -> IndexEntry:
        return self._index_entries[index]

    def iter(
        self,
        person_filter: Person = None,
        person_shift_filter: PersonShift = None,
        day_filter: date = None,
        day_shift_filter: Shift = None,
    ) -> Generator[IndexEntry, None, None]:
        """Return all indices that match the given filters

        This implementation is not efficient as it sorts and traverses all the indices every
        time. This should be fine for normal use as the cost of solving for the rota should be
        the bottleneck as the indices grow, not this.
        """

        if day_shift_filter is not None:
            assert (
                day_filter is not None
            ), "day_shift_filter can only be used together with day_filter"

        def _filter(idx_):
            (person_idx, person_shift_idx, day_idx, day_shift_idx) = idx_

            if (
                person_filter is not None
                and self._person_indices[person_filter] != person_idx
            ):
                return False

            if (
                person_shift_filter is not None
                and person_shift_filter != person_shift_idx
            ):
                return False

            if day_filter is not None:
                if self._day_indices[day_filter] != day_idx:
                    return False
                if (
                    day_shift_filter is not None
                    and self._day_shift_indices[(day_filter, day_shift_filter)]
                    != day_shift_idx
                ):
                    return False

            return True

        for idx in sorted(self._index_entries.keys()):
            if _filter(idx):
                yield self._index_entries[idx]



from dataclasses import dataclass
from typing import Any, Dict, List, Tuple

from or_shifty.person import Person
from or_shifty.shift import AssignedShift, ShiftType


@dataclass(frozen=True)
class PastShiftOffset:
    person: Person
    shift_type: ShiftType
    offset: int

    @classmethod
    def from_json(cls, serialised: Dict[str, Any]) -> "PastShiftOffset":
        return PastShiftOffset(
            person=Person(name=serialised["person"]),
            shift_type=ShiftType.from_json(serialised["shift_type"]),
            offset=int(serialised["offset"]),
        )


@dataclass(frozen=True)
class History:
    past_shifts: Tuple[AssignedShift, ...]
    offsets: Tuple[PastShiftOffset, ...]

    @classmethod
    def build(
        cls, past_shifts: List[AssignedShift] = (), offsets: List[PastShiftOffset] = ()
    ):
        past_shifts = sorted(past_shifts, key=lambda ps: ps.day, reverse=True)
        return cls(past_shifts=tuple(past_shifts), offsets=tuple(offsets))



import logging
from typing import List

from ortools.sat.python import cp_model
from ortools.sat.python.cp_model import INFEASIBLE

from or_shifty.config import Config
from or_shifty.constraints import (
    EVALUATION_CONSTRAINT,
    FIXED_CONSTRAINTS,
    Constraint,
)
from or_shifty.objective import Objective, RankingWeight
from or_shifty.shift import AssignedShift

log = logging.getLogger(__name__)


class Infeasible(Exception):
    pass


def solve(
    config: Config,
    objective: Objective = RankingWeight(),
    constraints: List[Constraint] = tuple(),
) -> List[AssignedShift]:
    constraints = _constraints(constraints)

    log.info(str(config.history_metrics))

    solver, assignments = _run_with_retries(config, objective, list(constraints))

    _validate_constraints_against_solution(solver, constraints, config, assignments)
    _display_objective_function_score(solver)

    solution = sorted(
        list(_solution(solver, config, assignments)), key=lambda s: (s.day, s.name)
    )
    log.info("Solution\n%s", "\n".join(f">>>> {shift}" for shift in solution))

    return solution


def evaluate(
    config: Config,
    objective: Objective,
    constraints: List[Constraint],
    solution: List[AssignedShift],
) -> None:
    constraints = _constraints(constraints)
    evaluation_constraint = EVALUATION_CONSTRAINT(priority=0, assigned_shifts=solution)

    log.info(str(config.history_metrics))

    solver, assignments = _run(config, objective, [evaluation_constraint])

    _validate_constraints_against_solution(solver, constraints, config, assignments)
    _display_objective_function_score(solver)

    solution = sorted(
        list(_solution(solver, config, assignments)), key=lambda s: (s.day, s.name)
    )
    log.info("Solution\n%s", "\n".join(f">>>> {shift}" for shift in solution))


def _constraints(constraints: List[Constraint]) -> List[Constraint]:
    constraints = list(constraints) + FIXED_CONSTRAINTS
    return sorted(constraints, key=lambda c: c.priority)


def _run_with_retries(config, objective, constraints):
    log.info("Running model...")
    while True:
        try:
            result = _run(config, objective, constraints)
            log.info("Solution found")
            return result
        except Infeasible:
            log.warning("Failed to find solution with current constraints")
            constraints = _drop_least_important_constraints(constraints)
            if constraints is None:
                raise
            log.info("Retrying model...")


def _drop_least_important_constraints(constraints):
    priority_to_drop = max(constraint.priority for constraint in constraints)
    if priority_to_drop == 0:
        return None
    log.debug("Dropping constraints with priority %s", priority_to_drop)
    constraints_to_drop = [
        constraint
        for constraint in constraints
        if constraint.priority == priority_to_drop
    ]
    log.info("Dropping constraints %s", ", ".join(str(c) for c in constraints_to_drop))

    return [
        constraint
        for constraint in constraints
        if constraint.priority != priority_to_drop
    ]


def _run(data, objective, constraints):
    model = cp_model.CpModel()

    assignments = init_assignments(model, data)

    for constraint in constraints:
        log.debug("Adding constraint %s", constraint)
        for expression, _ in constraint.generate(assignments, data):
            model.Add(expression)

    model.Maximize(objective.objective(assignments, data))

    solver = cp_model.CpSolver()
    status = solver.Solve(model)
    if status is INFEASIBLE:
        raise Infeasible()

    return solver, assignments


def init_assignments(model, data):
    assignments = {}
    for index in data.indexer.iter():
        assignments[index.idx] = model.NewBoolVar(
            f"shift_{index.person.name}_{index.person_shift}_{index.day}_{index.day_shift.name}"
        )
    return assignments


def _solution(solver, data, assignments):
    for day, day_shifts in data.shifts_by_day.items():
        for day_shift in day_shifts:
            for index in data.indexer.iter(day_filter=day, day_shift_filter=day_shift):
                if solver.Value(assignments[index.idx]) == 1:
                    yield index.day_shift.assign(index.person)


def _validate_constraints_against_solution(solver, constraints, data, assignments):
    for constraint in constraints:
        log.debug("Evaluating constraint %s against solution", constraint)
        for expression, impact in constraint.generate(assignments, data):
            expr = expression.Expression()
            bounds = expression.Bounds()
            value = solver.Value(expr)
            expr_valid = bounds[0] <= value <= bounds[1]
            if not expr_valid:
                log.debug(
                    "Solution violates constraint %s, expr %s, value %s, bounds %s, impact %s",
                    constraint,
                    expr,
                    value,
                    bounds,
                    impact,
                )
                log.warning("Solution violates constraint %s %s", constraint, impact)


def _display_objective_function_score(solver):
    log.info("Objective function score was %s", solver.ObjectiveValue())



from dataclasses import dataclass
from datetime import date
from typing import Dict, List

from or_shifty.history import History
from or_shifty.history_metrics import HistoryMetrics
from or_shifty.indexer import Indexer, PersonShift
from or_shifty.person import Person
from or_shifty.shift import Shift


@dataclass(frozen=True)
class Config:
    indexer: Indexer
    shifts_by_person: Dict[Person, List[PersonShift]]
    shifts_by_day: Dict[date, List[Shift]]
    max_shifts_per_person: int
    history: History
    history_metrics: HistoryMetrics
    now: date

    @classmethod
    def build(
        cls,
        people: List[Person],
        max_shifts_per_person: int,
        shifts_by_day: Dict[date, List[Shift]],
        history: History,
    ):
        now = min(shifts_by_day.keys())
        return cls(
            indexer=Indexer.build(people, max_shifts_per_person, shifts_by_day),
            shifts_by_person={
                person: [shift_idx for shift_idx in range(max_shifts_per_person)]
                for person in people
            },
            shifts_by_day=dict(shifts_by_day),
            max_shifts_per_person=max_shifts_per_person,
            history=history,
            history_metrics=HistoryMetrics.build(history, people, now),
            now=now,
        )



from abc import ABCMeta, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from datetime import date, datetime
from itertools import product
from typing import Dict, Generator, List, Optional, Set, Tuple

from ortools.sat.python.cp_model import IntVar, LinearExpr

from or_shifty.config import Config
from or_shifty.history_metrics import NEVER
from or_shifty.indexer import Idx
from or_shifty.person import Person
from or_shifty.shift import AssignedShift, Shift, ShiftType


@dataclass(frozen=True)
class ConstraintImpact:
    affected_person: Optional[Person]
    affected_day: Optional[date]

    def __str__(self):
        if self.affected_person is not None and self.affected_day is not None:
            return f"affecting {self.affected_person.name} on {self.affected_day}"
        if self.affected_person is not None:
            return f"affecting {self.affected_person.name}"
        if self.affected_day:
            return f"on {self.affected_day}"
        return None


class Constraint(metaclass=ABCMeta):
    def __init__(self, priority: int, name: Optional[str] = None):
        assert priority >= 0
        self._name = name or self.__class__.__name__
        self._priority = priority  # lower number is higher priority, 0 means this constraint is mandatory

    def __str__(self):
        return self._name

    @property
    def priority(self):
        return self._priority

    @abstractmethod
    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        yield from ()

    def __eq__(self, other):
        if other is None:
            return False
        if type(self) != type(other):
            return False
        if str(self) != str(other):
            return False
        return self.priority == other.priority


class EachDayShiftIsAssignedToExactlyOnePersonShift(Constraint):
    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        for day, day_shifts in data.shifts_by_day.items():
            for day_shift in day_shifts:
                yield (
                    (
                        sum(
                            assignments[index.idx]
                            for index in data.indexer.iter(
                                day_filter=day, day_shift_filter=day_shift
                            )
                        )
                        == 1
                    ),
                    ConstraintImpact(None, day),
                )


class EachPersonShiftIsAssignedToAtMostOneDayShift(Constraint):
    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, Person], None, None]:
        for person, person_shifts in data.shifts_by_person.items():
            for person_shift in person_shifts:
                yield (
                    (
                        sum(
                            assignments[index.idx]
                            for index in data.indexer.iter(
                                person_filter=person, person_shift_filter=person_shift
                            )
                        )
                        <= 1
                    ),
                    ConstraintImpact(person, None),
                )


class EachPersonsShiftsAreFilledInOrder(Constraint):
    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        for person, person_shifts in data.shifts_by_person.items():
            for person_shift_idx, person_shift in enumerate(person_shifts):
                shift_assigned = sum(
                    assignments[index.idx]
                    for index in data.indexer.iter(
                        person_filter=person, person_shift_filter=person_shift
                    )
                )
                next_idx = person_shift_idx + 1
                for subsequent_person_shift in person_shifts[next_idx:]:
                    subsequent_shift_assigned = sum(
                        assignments[index.idx]
                        for index in data.indexer.iter(
                            person_filter=person,
                            person_shift_filter=subsequent_person_shift,
                        )
                    )
                    yield (
                        shift_assigned >= subsequent_shift_assigned,
                        ConstraintImpact(person, None),
                    )


class EachPersonWorksAtMostXShiftsPerAssignmentPeriod(Constraint):
    def __init__(self, x=None, **kwargs):
        super().__init__(**kwargs)
        assert x is not None
        self._x = x

    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        assert (
            self._x <= data.max_shifts_per_person
        ), f"X in {self.__class__.__name__} must be <= than max_shifts_per_person"
        for person in data.shifts_by_person.keys():
            yield (
                (
                    sum(
                        assignments[index.idx]
                        for index in data.indexer.iter(person_filter=person)
                    )
                    <= 1
                ),
                ConstraintImpact(person, None),
            )

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        return self._x == other._x


class ThereShouldBeAtLeastXDaysBetweenOps(Constraint):
    def __init__(self, x=None, **kwargs):
        super().__init__(**kwargs)
        assert x is not None
        self._x = x

    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        for person, day in product(
            data.shifts_by_person.keys(), data.shifts_by_day.keys()
        ):
            date_last_on_shift = data.history_metrics.date_last_on_shift.get(person)

            if date_last_on_shift is None:
                continue

            if (day - date_last_on_shift).days > self._x:
                continue

            for index in data.indexer.iter(person_filter=person, day_filter=day):
                yield (
                    assignments[index.idx] == 0,
                    ConstraintImpact(person, day),
                )

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        return self._x == other._x


class ThereShouldBeAtLeastXDaysBetweenOpsOfShiftTypes(Constraint):
    def __init__(self, x=None, shift_types=None, **kwargs):
        super().__init__(**kwargs)
        assert x is not None
        assert shift_types is not None
        self._x = x
        self._shift_types = {
            ShiftType.from_json(shift_type) for shift_type in shift_types
        }

    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        for person, (day, day_shifts) in product(
            data.shifts_by_person.keys(), data.shifts_by_day.items()
        ):
            date_last_on_shift = max(
                data.history_metrics.date_last_on_shift_of_type.get(person, {}).get(
                    shift_type, NEVER
                )
                for shift_type in self._shift_types
            )

            if date_last_on_shift is None:
                continue

            if (day - date_last_on_shift).days > self._x:
                continue

            for day_shift in day_shifts:
                if day_shift.shift_type not in self._shift_types:
                    continue

                for index in data.indexer.iter(
                    person_filter=person, day_filter=day, day_shift_filter=day_shift
                ):
                    yield (
                        assignments[index.idx] == 0,
                        ConstraintImpact(person, day),
                    )

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        return self._x == other._x


class RespectPersonRestrictionsPerShiftType(Constraint):
    def __init__(self, forbidden_by_shift_type: Dict[str, List[str]] = None, **kwargs):
        super().__init__(**kwargs)
        assert forbidden_by_shift_type is not None
        self._forbidden_by_shift_type = {
            ShiftType.from_json(shift_type): set(names)
            for shift_type, names in forbidden_by_shift_type.items()
        }

    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        for person in data.shifts_by_person.keys():
            for day, day_shifts in data.shifts_by_day.items():
                for day_shift in day_shifts:
                    yield from self._generate_for_person_day_shift(
                        assignments, data, person, day, day_shift
                    )

    def _generate_for_person_day_shift(self, assignments, data, person, day, day_shift):
        if person.name not in self._forbidden_by_shift_type.get(
            day_shift.shift_type, set()
        ):
            return
        for index in data.indexer.iter(person_filter=person, day_filter=day):
            yield (assignments[index.idx] == 0, ConstraintImpact(person, day))

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        return self._forbidden_by_shift_type == other._forbidden_by_shift_type


class RespectPersonRestrictionsPerDay(Constraint):
    def __init__(self, restrictions: Dict[str, List[str]] = None, **kwargs):
        super().__init__(**kwargs)
        assert restrictions is not None
        self._restrictions = {
            person_name: {
                datetime.fromisoformat(weekday).date() for weekday in weekdays
            }
            for person_name, weekdays in restrictions.items()
        }

    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        for person, day in product(
            data.shifts_by_person.keys(), data.shifts_by_day.keys()
        ):
            if day in self._restrictions.get(person.name, set()):
                for index in data.indexer.iter(person_filter=person, day_filter=day):
                    yield (
                        assignments[index.idx] == 0,
                        ConstraintImpact(person, day),
                    )

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        return self._restrictions == other._restrictions


class PredeterminedAssignmentsConstraint(Constraint):
    """This constraint forces the solver to a given solution

    It is meant to be used with the evaluation mode so the solver can be forced to produce a given
    solution against which the constraints and objective function can then be evaluated.
    """

    def __init__(self, assigned_shifts: List[AssignedShift], **kwargs) -> None:
        super().__init__(**kwargs)
        self._assigned_shifts = assigned_shifts

    def generate(
        self, assignments: Dict[Idx, IntVar], data: Config
    ) -> Generator[Tuple[LinearExpr, ConstraintImpact], None, None]:
        selected_shifts = self._selected_shifts()

        for index in data.indexer.iter():
            key = (index.person, index.person_shift, index.day_shift)
            assignment = 1 if (key in selected_shifts) else 0
            yield (
                assignments[index.idx] == assignment,
                ConstraintImpact(None, None),
            )

    def _selected_shifts(self) -> Set[Tuple[Person, int, Shift]]:
        selected_shifts = set()
        next_person_shift = defaultdict(lambda: 0)

        for shift in self._assigned_shifts:
            selected_shifts.add(
                (shift.person, next_person_shift[shift.person], shift.unassigned())
            )
            next_person_shift[shift.person] += 1

        return selected_shifts

    def __eq__(self, other):
        if not super().__eq__(other):
            return False
        return self._assigned_shifts == other._assigned_shifts


FIXED_CONSTRAINTS = [
    EachDayShiftIsAssignedToExactlyOnePersonShift(priority=0),
    EachPersonShiftIsAssignedToAtMostOneDayShift(priority=0),
    EachPersonsShiftsAreFilledInOrder(priority=0),
]


CONSTRAINTS = {
    constraint.__name__: constraint
    for constraint in [
        EachPersonWorksAtMostXShiftsPerAssignmentPeriod,
        ThereShouldBeAtLeastXDaysBetweenOps,
        ThereShouldBeAtLeastXDaysBetweenOpsOfShiftTypes,
        RespectPersonRestrictionsPerShiftType,
        RespectPersonRestrictionsPerDay,
    ]
}

EVALUATION_CONSTRAINT = PredeterminedAssignmentsConstraint



import logging
import sys

from or_shifty.cli import Inputs, InvalidInputs, parse_args, write_output
from or_shifty.config import Config
from or_shifty.model import Infeasible, evaluate, solve

logging.basicConfig(
    stream=sys.stderr, level=logging.INFO, format="%(levelname)-7s - %(message)s",
)
log = logging.getLogger(__name__)


def main() -> None:
    inputs = parse_args()
    configure_logging(inputs.verbose)
    config = Config.build(
        people=inputs.people,
        max_shifts_per_person=inputs.max_shifts_per_person,
        shifts_by_day=inputs.shifts_by_day,
        history=inputs.history,
    )

    if inputs.evaluate:
        evaluation_mode(inputs, config)
    else:
        solving_mode(inputs, config)


def evaluation_mode(inputs: Inputs, config: Config) -> None:
    try:
        evaluate(
            config=config,
            objective=inputs.objective,
            constraints=inputs.constraints,
            solution=inputs.output,
        )
    except Infeasible:
        log.error("The provided output is infeasible for the solver")
        exit(1)
    except InvalidInputs as e:
        log.error(e.msg)
        exit(1)


def solving_mode(inputs: Inputs, config: Config) -> None:
    try:
        solution = solve(
            config=config, objective=inputs.objective, constraints=inputs.constraints,
        )
    except Infeasible:
        log.error("Unable to solve for the given constraints")
        exit(1)
    else:
        if inputs.output_path is not None:
            write_output(inputs.output_path, solution)


def configure_logging(verbose=False):
    logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)



import argparse
import json
import logging
from dataclasses import dataclass
from datetime import date, datetime
from typing import Dict, List, Optional

from or_shifty.constraints import CONSTRAINTS, Constraint
from or_shifty.history import History, PastShiftOffset
from or_shifty.objective import OBJECTIVE_FUNCTIONS, Objective
from or_shifty.person import Person
from or_shifty.shift import AssignedShift, Shift, ShiftType

log = logging.getLogger(__name__)


class InvalidInputs(Exception):
    def __init__(self, msg):
        self.msg = msg


@dataclass(frozen=True)
class Inputs:
    people: List[Person]
    max_shifts_per_person: int
    shifts_by_day: Dict[date, List[Shift]]
    objective: Objective
    constraints: List[Constraint]
    history: History
    verbose: bool
    output_path: Optional[str]
    evaluate: bool
    output: Optional[List[AssignedShift]]


def parse_args(args=None) -> Inputs:
    parser = argparse.ArgumentParser(
        description="Automatic ops shift allocator using constraint solver"
    )
    parser.add_argument(
        "--config",
        dest="config",
        action="store",
        required=True,
        help="Path to json file contain the application config",
    )
    parser.add_argument(
        "--history",
        dest="history",
        action="store",
        required=True,
        help="Path to json file containing history of past shifts",
    )
    parser.add_argument(
        "-v",
        dest="verbose",
        action="store_true",
        default=False,
        help="Change the log level to debug",
    )
    parser.add_argument(
        "--output",
        dest="output",
        action="store",
        default=None,
        help="Path to file in which to write the solution",
    )
    parser.add_argument(
        "--evaluate",
        dest="evaluate",
        action="store_true",
        default=False,
        help="If selected then instead of running the solver, an existing solution will "
        "be evaluated against the given config and history and the script will print "
        "the score of the objective function and any violated constraints. When used "
        "--output must also be provided",
    )

    parsed_args = parser.parse_args(args)

    return _parse_inputs(
        config_path=parsed_args.config,
        history_path=parsed_args.history,
        verbose=parsed_args.verbose,
        output_path=parsed_args.output,
        evaluate=parsed_args.evaluate,
    )


def _parse_inputs(
    config_path: str,
    history_path: str,
    verbose: bool,
    output_path: Optional[str],
    evaluate: bool,
) -> Inputs:
    _validate_args(output_path, evaluate)

    with open(config_path, "r") as f:
        config = json.load(f)

    with open(history_path, "r") as f:
        history = json.load(f)

    shifts_by_day = _parse_shifts_by_day(config)

    if evaluate:
        output = read_output(output_path)
        _validate_evaluation_output(shifts_by_day, output)
    else:
        output = None

    return Inputs(
        people=_parse_people(config),
        max_shifts_per_person=_parse_max_shifts_per_person(config),
        shifts_by_day=shifts_by_day,
        objective=_parse_objective(config),
        constraints=_parse_constraints(config),
        history=_parse_history(history),
        verbose=verbose,
        output_path=output_path,
        evaluate=evaluate,
        output=output,
    )


def _validate_args(output: Optional[str], evaluate: bool,) -> None:
    if evaluate and output is None:
        raise InvalidInputs("When in evaluate mode output path must be provided")


def _validate_evaluation_output(
    shifts_by_day: Dict[date, List[Shift]], output: List[AssignedShift]
) -> None:
    expected_shifts = {shift for shifts in shifts_by_day.values() for shift in shifts}
    output_shifts = {shift.unassigned() for shift in output}
    missing_from_output = output_shifts - expected_shifts
    extra_in_output = expected_shifts - output_shifts
    if missing_from_output or extra_in_output:
        raise InvalidInputs(
            f"There is a discrepancy between the shifts in the output and those from config. "
            f"Extra in output: {extra_in_output}, Missing from output: {missing_from_output}"
        )


def _parse_people(config) -> List[Person]:
    return [Person(name=person["name"]) for person in config["people"]]


def _parse_max_shifts_per_person(config) -> int:
    return int(config["max_shifts_per_person"])


def _parse_shifts_by_day(config) -> Dict[date, List[Shift]]:
    shifts = {}

    for shift in config["shifts"]:
        day = datetime.fromisoformat(shift["day"]).date()
        name = shift["name"]
        shift_type = ShiftType.from_json(shift["type"])

        day_shift = Shift(day=day, name=name, shift_type=shift_type)

        shifts.setdefault(day, []).append(day_shift)

    return shifts


def _parse_objective(config) -> Objective:
    return OBJECTIVE_FUNCTIONS[config["objective"]]()


def _parse_constraints(config) -> List[Constraint]:
    return [
        CONSTRAINTS[constraint["type"]](
            priority=constraint["priority"],
            name=constraint.get("name"),
            **constraint.get("params", {}),
        )
        for constraint in config["constraints"]
    ]


def _parse_history(history) -> History:
    offsets = []
    shifts = []

    for offset in history["offsets"]:
        offsets.append(PastShiftOffset.from_json(offset))

    for shift in history["shifts"]:
        shifts.append(AssignedShift.from_json(shift))

    return History.build(past_shifts=shifts, offsets=offsets)


def read_output(output_path: str) -> List[AssignedShift]:
    with open(output_path, "r") as f:
        output = json.load(f)
    shifts = [AssignedShift.from_json(shift) for shift in output["shifts"]]
    return shifts


def write_output(output_path: str, solution: List[AssignedShift]):
    log.info("Writing solution to %s...", output_path)
    solution_json = {
        "shifts": [assigned_shift.to_json() for assigned_shift in solution]
    }
    with open(output_path, "w") as f:
        json.dump(solution_json, f, indent=2)
    log.info("Solution written successfully")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['or_shifty']

package_data = \
{'': ['*']}

install_requires = \
['ortools>=7.2,<8.0']

entry_points = \
{'console_scripts': ['shifty = or_shifty.app:main']}

setup_kwargs = {
    'name': 'or-shifty',
    'version': '1.1.0',
    'description': '',
    'long_description': '# OR-Shifty\n`OR-Shifty` is a CLI app to automate assignment of on-call shifts via constraint solving by use of the\n[OR-tools](https://github.com/google/or-tools) library.\n\nIt is designed to accept two JSON files as input:\n\n- a config file which describes the desired on-call rota, the people to assign, and any constraints\n- a history file which describes past shifts per person, to be used by certain constraints described below\n\nIt will then attempt to create a rota allocation given these inputs and print it either on terminal or as JSON to an\noutput file if provided.\n\nThe given constraints can be prioritised so if a solution cannot be found that satisfies all of them the tool will\ndrop the least important constraint and try again, repeatedly, until a solution is found or no more constraints can\nbe dropped. If any constraints had to be dropped the tool will also print to the terminal which of the original\nconstraints were violated, for which person, and on which day.\n\n## Installation\nYou will need Python 3.7 or greater.\n\n```bash\npip install or-shifty\n```\n\n## Use\nHelp:\n\n```bash\nshifty --help\n```\n\n### Solver mode\nFor using shifty in its default solver mode.\n\nAn output file path can optionally be provided in which case shifty will write the solution to the file in JSON\nformat. If a file already exists at that path then it will overwritten.\n\n```bash\nshifty \\\n    --config <path_to_config.json> \\\n    --history <path_to_history.json> \\\n    [--output <path_to_optional_output.json>]\n```\n\n### Evaluation mode\nShifty can also be run in evaluation mode.\n\nIn this mode an output with an existing solution must be provided. Shifty will then evaluate this solution against\nthe constraints, objective function, and history provided and it will print which (if any) constraints are violated\nby the solution and the score of the objective function. This is intended as a diagnostic tool to inspect shifty\'s\ndecisions.\n\nWhen in this mode the output file must contain exactly an assigned shift for every shift specified in config otherwise\nshifty will exit with an error.\n\n```bash\nshifty \\\n    --config <path_to_config.json> \\\n    --history <path_to_history.json> \\\n    --output <path_to_optional_output.json> \\\n    --evaluate\n```\n\n### Shift types\nAny shift can be assigned one of three types. These types have no intrinsic meaning and only exist so different \nconstraints can be applied to different shifts, e.g. it might be desired to have more unassigned days after having a\nSaturday shift or similar.\n\nThe currently supported shift types are:\n\n- STANDARD\n- SPECIAL_A\n- SPECIAL_B\n\n### Constraints\nSolver constraints fall into one of two categories:\n\n- Mandatory, that are always included\n- User selected, that can be included and configured using the config input file\n\nFor each user selected constraint there will be an example of the JSON configuration needed to include it.\n\nUser selected constraints can also be given a priority. If a solution can not be found the constraints with the highest\npriority number will be dropped and the solver will retry, then the constraints with the second highest priority\nnumber will be dropped and so on. Priority must be a positive integer. A priority of 0 means the constraint cannot ever\nbe dropped. If multiple constraints have the same priority and that priority is due to be dropped then all these\nconstraints will be dropped together.\n\nConstraints can also be named to make them easier to manager by providing `"name": "constraint name"` in the JSON\nconfig.\n\nIt is valid to include the same type of constraint multiple times in the config, e.g. by having a stricter constraint\nwith a higher priority number, followed by a more permissive one with a lower priority number. This way the solver\nwill try to meet the strict one first, but if it cannot it will then at least try to meet the more permissive one.\n\nThe mandatory constraints are mostly common sense and can be seen in the code, e.g. you can\'t assign Eve\'s first shift\ntwice, etc.\n\nThe user selected constraints are as follows:\n\n#### Each person works at most X shifts per assignment period\n```json\n"constraints": [\n    {\n      "type": "EachPersonWorksAtMostXShiftsPerAssignmentPeriod",\n      "priority": 0,\n      "params": {"x":  1}\n    }\n]\n```\n\nAn assignment period is one run of the program with given input files. This constraints that a person can only be\nassigned X of these shifts.\n\n#### There should be at least X days between ops\n```json\n"constraints": [\n    {\n      "type": "ThereShouldBeAtLeastXDaysBetweenOps",\n      "priority": 0,\n      "params": {"x":  4}\n    }\n]\n```\n\nSubtract the last date on ops from the date assigned on a shift. The resulting number of days must be greater that X.\n\n#### There should be at least X days between ops of shift types\n```json\n"constraints": [\n    {\n      "type": "ThereShouldBeAtLeastXDaysBetweenOpsOfShiftTypes",\n      "priority": 0,\n      "params": {"x":  8, "shift_types": ["SPECIAL_A", "SPECIAL_B"]}\n    }\n]\n```\nSubtract the last date on ops of any the given types from the date assigned on a shift of one of the given types. The\nresulting number of days must be grater than X. In the example configuration someone cannot be assigned to a\nSPECIAL_A or SPECIAL_B type shift if they have had a SPECIAL_A or SPECIAL_B shift in the last 8 days.\n\n#### Respect person restrictions per shift type\n```json\n"constraints": [\n    {\n      "type": "RespectPersonRestrictionsPerShiftType",\n      "priority": 0,\n      "params": {"forbidden_by_shift_type": {"special_a": ["Alice"]}}\n    }\n]\n```\n\nDisallow particular people from being assigned any shifts of a given type.\n\n#### Respect person restrictions per day\n```json\n"constraints": [\n    {\n      "type": "RespectPersonRestrictionsPerDay",\n      "name": "Holidays",\n      "priority": 0,\n      "params": {"restrictions": {"Alice": ["2019-11-01"]}}\n    }\n]\n```\n\nDisallow particular people from being assigned any shifts on a given date.\n\n### Objective function\nWhen scheduling the solver will choose between possible solutions that satisfy the constraints by using an objective \nfunction. Each possible solution will be assigned a score using this function and the solution with the highest score\nwill be chosen.\n\nThere is currently support for only one type of objective function called `RankingWeight`.\n\n#### RankingWeight\nThis objective function will rank people for each shift type based on the following criteria:\n\n1. How many shifts of the given type have they done before\n1. How recently have they been on shift\n1. Their name (just to ensure there is a total ordering)\n\nIn addition it will also rank someone being assigned a second shift below everyone else being assigned their first shift\nand so on for possible additional shifts on the same person.\n\ne.g. we have Bob, Alice, and Eve. Bob has done 10 shifts, Alice and Eve have done 8 each. Alice was last on call 2 days\nago and Even was 4 days ago. Everyone can be assigned a maximum of two shifts. The ranking would be:\n\n1. Eve, 1st shift\n1. Alice, 1st shift\n1. Bob, 1st shift\n1. Eve, 2nd shift\n1. Alice, 2nd shift\n1. Bob, 2nd shift\n\nEve is given the highest priority because she is tied for least number of shifts done, but has not been on call as\nrecently. Second shifts are ranked the same way, but after all first shifts.\n\nEach level of this rank is then assigned a weight that is higher the further up in the ranking. \n\nThere is a large gap in the weights when switching from the first shift to the second and so on in order to discourage\nassigning multiple shifts before exhausting other possibilities.\n\nThis process is repeated for each shift type. The resulting weights are then added together to produce the final score.\nThis final score is then given as a reward to the solver for assigning the corresponding person/shift.\n\n### Config\nExamples can be found under `examples`.\n\nThe config file describes the desired on-call rota, the people to assign, and any constraints. The meaning of its\nsections are as follows:\n\n- `shifts`: contains one entry for each shift to assign, along with its name, and type\n- `people`: the people to assign\n- `max_shifts_per_person`: a hard limit on how many shifts a single person can be assigned\n- `objective`: the chosen objective function\n- `constraints`: the chosen constraints\n\n### History\nExamples can be found under `examples`.\n\nThe history file describes past shifts per person, to be used by certain constraints. The meaning of its sections\nare as follows:\n\n- `shifts`: one entry per past shift a person has done, including its type\n- `offset`: offsets to be added to number of shifts done per type for each person\n\n## Development\nOR-Shifty is Python3 application developed using [Poetry](https://github.com/python-poetry/poetry). The minimum required\nPython version, project dependencies, and other project information can be found in the `pyproject.toml` file.\n\nFor ease of development there is a `Makefile`.\n\n```bash\nmake help     # Get a list of all supported make targets\nmake freeze   # Update and freeze python requirements\nmake test     # Run tests\nmake format   # Run formatters and linters\nmake verify   # Run formatters and linters in check only mode\nmake install  # Install project dependencies from poetry.lock, project module, and `shifty` script\nmake build    # Build source and wheels\n```\n\nBefore submitting any pull requests `make test` and `make verify` must both be run an be passing.\n\n## License\nSee the LICENSE file that is included with this repository.\n',
    'author': 'Thomas Maroulis',
    'author_email': 'thomasmaroulis@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Dalamar42/or-shifty',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
